/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <iostream>

#include "u2d_mesh.h"
#include "boost/numeric/mtl/mtl.hpp"

using namespace std;
using namespace mtl;

u2d_Mesh::u2d_Mesh(int _spatial_dim,
        int _d_nodes_n, int _b_nodes_n,
        int _d_elem_n, int _b_elem_n) :
spatial_dim(_spatial_dim),
d_nodes_n(_d_nodes_n), b_nodes_n(_b_nodes_n),
d_elem_n(_d_elem_n), b_elem_n(_b_elem_n),
reverse(false),
rr(_spatial_dim, _d_nodes_n),
jjd(DOMAIN_ELEMENT_SIZE, _d_elem_n),
jjb(BOUNDARY_ELEMENT_SIZE, _b_elem_n),
neigh(DOMAIN_ELEMENT_SIZE, _d_elem_n),
jjb_s(BOUNDARY_ELEMENT_SIZE, _b_elem_n),
neigh_s(_b_elem_n, 0),
b_id(_b_elem_n, 0),
jjb_sd(_b_nodes_n, 0) {

};

//u2d_Mesh::u2d_Mesh(u2d_Mesh& orig) :
//spatial_dim(orig.getSpatialDimension()),
//d_nodes_n(orig.getDomainNodeNumber()),
//b_nodes_n(orig.getBoundaryNodeNumber()),
//d_elem_n(orig.getDomainElementNumber()),
//b_elem_n(orig.getBoundaryElementNumber()),
//reverse(orig.getReverse()) {
//
//    reverse = orig.getReverse();
//    rr = orig.rr;
//    jjd = orig.jjd;
//    jjb = orig.jjb;
//    neigh = orig.neigh;
//    jjb_s = orig.jjb_s;
//    neigh_s = orig.neigh_s;
//    b_id = orig.b_id;
//    jjb_sd = orig.jjb_sd;
//
//}

int u2d_Mesh::getDomainElementNumber() {
    return d_elem_n;
}

int u2d_Mesh::getBoundaryElementNumber() {
    return b_elem_n;
}

int u2d_Mesh::getDomainNodeNumber() {
    return d_nodes_n;
}

int u2d_Mesh::getBoundaryNodeNumber() {
    return b_nodes_n;
}

int u2d_Mesh::getSpatialDimension() {
    return spatial_dim;
}

bool u2d_Mesh::getReverse() {
    return reverse;
}

void u2d_Mesh::resize(int _spatial_dim,
        int _d_nodes_n, int _b_nodes_n,
        int _d_elem_n, int _b_elem_n) {
    rr.change_dim(_spatial_dim, _d_nodes_n);
    jjd.change_dim(DOMAIN_ELEMENT_SIZE, _d_elem_n);
    jjb.change_dim(BOUNDARY_ELEMENT_SIZE, _b_elem_n);
    jjb_s.change_dim(BOUNDARY_ELEMENT_SIZE, _b_elem_n);
    neigh.change_dim(DOMAIN_ELEMENT_SIZE, _d_elem_n);
    neigh_s.change_dim(_b_elem_n);
    b_id.change_dim(_b_elem_n);
    jjb_sd.change_dim(_b_nodes_n);

    spatial_dim = _spatial_dim;
    d_nodes_n = _d_nodes_n;
    b_nodes_n = _b_nodes_n;
    d_elem_n = _d_elem_n;
    b_elem_n = _b_elem_n;
}

void u2d_Mesh::setBoundaryNodeReversal(bool _reverse) {
    reverse = _reverse;
}
/**
 * This calculates the connectivity for boundary elements from domain elements,
 * filling the jjb matrix with correct values. If reversal of boundary elements
 * nodes has been requested, this function will also reverse them.
 * @remarks This must be called after the matrix has been created.
 */
void u2d_Mesh::getReady() {

    /* Reverse the boundary nodes if so has been asked */
    if(reverse) {
        for(int i = 0; i < getBoundaryElementNumber(); i++ ) {
            int tmp_id;
            tmp_id = jjb_s(1, i);
            jjb_s(1, i) = jjb_s(0, i);
            jjb_s(0, i) = tmp_id;
        }
    }

    /* ORIGINAL CODE, down there is my improved version
     * The extra exit condition here is also mine
    for (int i = 0; i < b_elem_n; i++) {
        for (int n = 0; n < BOUNDARY_ELEMENT_SIZE; n++) {
            for (int j = 1; j <= d_nodes_n; j++) {
                // The -1 is to fix indexing of Fortran vs C
                // Fortran starts from 1, while C from 0
                //
                if (jjb_sd(jjb_s(n, i) - 1) == j) {
                    jjb(n, i) = j;
                    break; // j = d_nodes_n;  We found the node, we stop looking for it
                }
            }
        }
    }
     */

    /* My improved version of the code */
    for (int i = 0; i < b_elem_n; i++) {
        for (int n = 0; n < BOUNDARY_ELEMENT_SIZE; n++) {
            jjb(n, i) = jjb_sd(jjb_s(n, i) - 1);
        }
    }
    /* Lovely debug output
    for (int k; k < b_elem_n; k++) {
        cout << jjb(0, k) << " " << jjb(1, k) << endl;
    }
     * */
}

/**
 * Return the nodes of a given domain element identified by it's id.
 * Calculate the x and y vectors of the i-th element
 * like: element 1, coordinate x of it's first
 * node AND coordinate x of the second node AND
 * coorinate x of the third node.
 *              NODES ID
 * element (i-th) 341 -> | 125 | -> | 0.645  0.454543 |
 *                       | 435 |    | 0.122  0.54564  |
 *                       | 673 |    | 0.454  0.6454   |
 *                                    ^^^^^^^^^^^^^^^
 *                                   return this matrix
 */
multi_vector<dense_vector<double> > u2d_Mesh::getDomainElementNodes(int elem_id) {

    multi_vector<dense_vector<double> > elem(DOMAIN_ELEMENT_SIZE, getSpatialDimension());
    for (int ds = 0; ds < DOMAIN_ELEMENT_SIZE; ds++) {
        elem(ds, 0) = rr(0, jjd(ds, elem_id - 1) - 1); // X of current row
        elem(ds, 1) = rr(1, jjd(ds, elem_id - 1) - 1); // Y of current row
    }
    return elem;
}

/**
 * Return the nodes of a given boundary element identified by it's id.
 * Calculate the x and y vectors of the i-th element
 * like: element 1, coordinate x and y  of it's first
 * node AND coordinate x and y of the second node AND
 * coorinate x and y of the third node.
 *         ELEMENT ID    NODEs IDs       COORDINATES
 * element (i-th) 341 -> | 125 | -> | 0.645  0.454543 |
 *                       | 435 |    | 0.122  0.54564  |
 *                                    ^^^^^^^^^^^^^^^
 *                                   return this matrix
 */
multi_vector<dense_vector<double> > u2d_Mesh::getBoundaryElementNodes(int elem_id) {

    multi_vector<dense_vector<double> > elem(BOUNDARY_ELEMENT_SIZE, getSpatialDimension());
    for (int ds = 0; ds < BOUNDARY_ELEMENT_SIZE; ds++) {
        elem(ds, 0) = rr(0, jjb(ds, elem_id - 1) - 1); // X of current row
        elem(ds, 1) = rr(1, jjb(ds, elem_id - 1) - 1); // Y of current row
    }
    return elem;
}

/**
 * Return the nodes id of a given domain element identified by it's element id.
 *         ELEMENT ID    NODEs IDs
 * element (i-th) 341 -> | 125 |
 *                       | 435 |
 *                       | 673 |
 *                     ^^^^^^^^^^^^^
 *                   return this vector
 */
dense_vector<double> u2d_Mesh::getDomainElementNodesId(int elem_id) {
        dense_vector<double> elem(3, 0.0);
        // The next one stopped working for no know reason
        //elem = jjd[mtl::iall][elem_id];
        // The next one works instead
        elem = jjd[mtl::irange()][elem_id];
        // The old manual way
        //elem(0) = jjd[0][elem_id];
        //elem(1) = jjd[1][elem_id];
        //elem(2) = jjd[2][elem_id];
        return elem;
}

/**
 * Returns the side number of a boundary element identified by it's id. The side
 * number is the side the element is part of.
 *             side 4
 *             +--2-+
 *             |   /|
 * side 1      1  5 |  side 3
 *             | /  3
 *             |/   |
 *             +--4-+
 *             side 2
 * So given this example the function would return:
 * getElementSideId(1) ---> 1
 * getElementSideId(2) ---> 4
 * getElementSideId(3) ---> 3
 * getElementSideId(4) ---> 2
 * getElementSideId(5) ---> NOT A BOUNDARY ELEMENT, mtl exception!
 *
 */
int u2d_Mesh::getElementSideId(int jjb_id) {
    return b_id(jjb_id - 1);
}