/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */


#ifndef U2D_MESH_H
#define	U2D_MESH_H

#include <iostream>

#include "boost/numeric/mtl/mtl.hpp"

#define NODE_SIZE 2  /* We are in 2d so nodes have two coordinates, that are
                      * double numbers.
                      */
#define DOMAIN_ELEMENT_SIZE 3 /* An element of the domain is made of three nodes
                               * since we work with triangles, so we need a
                               * triplet of integer numbers to represent it.
                               */
#define BOUNDARY_ELEMENT_SIZE 2 /* An element of the boundary is made of two nodes
                               * since it's a straight line, so we need a
                               * couple of integer numbers to represent it.
                               */
#define ZERO_SIZE 0 /* We allocate all matrices and vectors with dummy size
                      * and change_dim them when we discover the correct size */

using namespace mtl;

/**
 * The Triplet jjb_s, neigh_s and b_id are intended to be placed side by
 * side they have same order so first element of one corresponds to first
 * element of the other. They include all info of the boundary numbered by
 * boundary.
 * The vector jjb_sd contain the surface->domain (sd, boundary->domain)
 * correspondence of IDs, so that each element of jjb_sd contain the id of the
 * corresponding domain node. Example: Node 15 in the numeration of boundary
 * nodes correspond to node 1456 in the domain numeration, so position 15 of
 * jjb_sd will be == 1456.
 *
 */
class u2d_Mesh {
private:
    int spatial_dim;
    int d_nodes_n;
    int b_nodes_n;
    int d_elem_n;
    int b_elem_n;
    bool reverse;

public:
    //u2d_Mesh();
    u2d_Mesh(int _spatial_dim = 0,
            int _d_nodes_n = 0, int _b_nodes_n = 0,
            int _d_elem_n = 0, int _b_elem_n = 0);
    //u2d_Mesh(u2d_Mesh& orig);
    int getDomainElementNumber();
    int getBoundaryElementNumber();
    int getDomainNodeNumber();
    int getBoundaryNodeNumber();
    int getSpatialDimension();
    bool getReverse();
    void resize(int _spatial_dim, int _d_nodes_n, int _b_nodes_n, int _d_elem_n, int _b_elem_n);
    void getReady();
    void setBoundaryNodeReversal(bool _reverse);
    
    dense2D<double> rr; /* Coordinates of nodes */
    dense2D<int> jjd; /* Triplets of node's ID of each domain element.
                          * One for each column */
    dense2D<int> jjb; /* Couples of node's ID of each boundary element.
                          * One for each column */
    dense2D<int> neigh; /* Elements that surround the element.
                            * One for each column */
    dense2D<int> jjb_s; /* Couples of node's ID of each boundary element.
                            * NOT NUMBERED by domain nodes id, BUT by boundary
                            * nodes id (as in neighs). One couple for each column */
    dense_vector<int> neigh_s; /* Numeration of the boundary elements, same order
                                  * of b_id vector. */
    dense_vector<int> b_id; /* Id of the boundary (the side) the boundary
                                * element is part of. */
    dense_vector<int> jjb_sd; /* Correspondence of the boundary IDs with the
                                  * domain IDs, so each position of this vector
                                  * contains the ID of the corresponding node in
                                  * domain node numeration. */
    multi_vector<dense_vector<double> > getDomainElementNodes(int elem_id);
    multi_vector<dense_vector<double> > getBoundaryElementNodes(int elem_id);
    dense_vector<double> getDomainElementNodesId(int elem_id);
    int getElementSideId(int jjb_id);



};




#endif	/* U2D_MESH_H */

