/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2008 Gabriele Dini Ciacci
  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <string>
#include <string.h>
#include <stdio.h>

#include "u2d_umesh2d_output.h"
#include "exceptions.h"
#include "u2d_core.h"

using std::string;

/***********************************/
/**********  PUBLIC API ************/
/***********************************/

u2d_Umesh2dOutput::u2d_Umesh2dOutput() {
  firstprimitive = NULL;
  lastprimitive = NULL;
  elements = 0;
  vertexs = 0;
  output_format = 0;
  steiners = NULL;
  backgrids = NULL;
  steiner_n = 0;
  backgrid_n = 0;
}

void u2d_Umesh2dOutput::select_prim_list(Primitive *first, Primitive *last, int prim_number, int vertex_n){
  firstprimitive = first;
  lastprimitive = last;
  elements = prim_number;
  vertexs = vertex_n;
}

void u2d_Umesh2dOutput::set_steiners(steiner *steiners_, int number){
  steiners = steiners_;
  steiner_n = number;
  printf("steiner_n %i\n", steiner_n);
}

void u2d_Umesh2dOutput::set_backgrids(backgrid *backgrid_, int number){
  backgrids = backgrid_;
  backgrid_n = number;
  printf("backgrid_n %i\n", backgrid_n);
}

void u2d_Umesh2dOutput::output()
    throw (u2d_IO_Exception&) {
    try {
        output_boundary_file();
        output_domain_file();
    }
    catch (...) {
        throw;
    }
}

void u2d_Umesh2dOutput::setFilename(string filename)
    throw (u2d_Parse_Error_Exception &) {
    if(filename.empty() == true) {
        throw new u2d_Parse_Error_Exception("Specified project name is empty");
    }
    project_name = filename;
    try {
        set_boundary_filename();
        set_domain_filename();
    }
    catch (u2d_Parse_Error_Exception &) {
        throw;
    }
}

string u2d_Umesh2dOutput::getFilename() {
    return project_name;
}

void u2d_Umesh2dOutput::setFormat(int format) {
  output_format = format;
}

/***********************************/
/**********  PRIVATE API ***********/
/***********************************/


/* FILE OUTPUT */

void u2d_Umesh2dOutput::open_boundary_file()
    throw (u2d_IO_Exception &) {
    boundary_file = fopen(boundary_filename.c_str(), "w+");
    if(boundary_file == NULL) {
        throw new u2d_IO_Exception("Can't open boundary file\n");
    }
}

void u2d_Umesh2dOutput::open_domain_file()
    throw (u2d_IO_Exception &) {
    domain_file = fopen(domain_filename.c_str(), "w+");
    if(domain_file == NULL) {
        throw new u2d_IO_Exception("Can't open domain file\n");
    }
}

void u2d_Umesh2dOutput::set_boundary_filename() {
  size_t position;
  if((position = project_name.find_last_of('/', project_name.at(0))) != string::npos) {
      position++; // we move after the /
  }
  else {
      position = 0;
  }
  boundary_filename = project_name;
  boundary_filename.insert(position, "boundary.");
}

void u2d_Umesh2dOutput::set_domain_filename() {
  size_t position;
  if((position = project_name.find_last_of('/', project_name.at(0))) != string::npos) {
      position++; // we move after the /
  }
  else {
      position = 0;
  }
  domain_filename = project_name;
  domain_filename.insert(position, "domain.");
}

void u2d_Umesh2dOutput::output_boundary_file()
    throw (u2d_IO_Exception &) {
    if(firstprimitive == NULL || elements == 0 || lastprimitive == NULL){
        throw new u2d_IO_Exception("output_boundary_file: Primitive list is "\
                                   "has not been initialized propely "\
                                   "or is empty.\n");
    }
    if(boundary_filename.empty() == true) {
        throw new u2d_IO_Exception("output_boundary_file: Invalid filename\n");
    }
    try {
        open_boundary_file();
    }
    catch (u2d_IO_Exception &) {
        throw;
    }

    /* Print the file */
    printformat(boundary_file, output_format);
    printnvne(boundary_file);
    printverelem(boundary_file);
    printgeom(boundary_file);
    printpriminfo(boundary_file);
    printtopo(boundary_file);
    printprimdata(boundary_file);
    fclose(boundary_file);
}

void u2d_Umesh2dOutput::output_domain_file()
    throw (u2d_IO_Exception &) {
    if(domain_filename.empty() == true) {
        throw new u2d_IO_Exception("output_boundary_file: Invalid filename\n");
    }
    try {
        open_domain_file();
    }
    catch (u2d_IO_Exception &) {
        throw;
    }

    fprintf(domain_file,\
            "    NSTEP    LSMOOTH(1/0)\n"\
            "    1000    1\n");

    printhash(domain_file);

    fprintf(domain_file,\
            "    N	(BACKGRID POINTS)\n"\
            "    %i\n"\
            "     X        Y	   HDATA\n",
            backgrid_n);
    for(int i = 0; i < backgrid_n; i++){
        fprintf(domain_file, "     %.2f     %.2f     %.2f\n",
                backgrids[i].x, backgrids[i].y, backgrids[i].hdata);
    }
	   
    printhash(domain_file);

    fprintf(domain_file,\
            "    N	(STEINER POINTS)\n"\
            "    %i\n"\
            "     X        Y	    \n",
            steiner_n);
    for(int i = 0; i < steiner_n; i++){
        fprintf(domain_file, "     %.2f     %.2f\n",
                steiners[i].x, steiners[i].y);
    }
    fclose(domain_file);
}

/* FILE GENERATION */

inline void u2d_Umesh2dOutput::printhash(FILE *output) {
  fprintf(output, "#\n");
}

inline void u2d_Umesh2dOutput::printformat(FILE *output, int format){
  switch(format) {
  case FORMAT_UMESH2D :
    break;
  case FORMAT_UMESH2D_FE :
    fprintf(output, "#\n" \
	    "######## CHOICE BETWEEN FINITE ELEMENTS" \
	    " vs VOLUME BOUNDARY NODES ########\n" \
	    "#\n"\
	    );
    fprintf(output, "FE\n");
    break;
  case FORMAT_UMESH2D_FV :
    fprintf(output, "#\n" \
	    "######## CHOICE BETWEEN FINITE ELEMENTS" \
	    " vs VOLUME BOUNDARY NODES ########\n" \
	    "#\n"\
	    );
    fprintf(output, "FV\n");
    break;
  }
}

inline void u2d_Umesh2dOutput::printgeom(FILE *output) {
  fprintf(output, "#\n" \
                  "######## GEOMETRY ########\n" \
	          "#\n"\
	  );
}

inline void u2d_Umesh2dOutput::printtopo(FILE *output) {
  fprintf(output, "#\n" \
                  "######## TOPOLOGY AND METRIC ########\n" \
	          "#\n"\
	  );
}

inline void u2d_Umesh2dOutput::printnvne(FILE *output) {
  fprintf(output, "#\n" \
                  "######## GLOBAL DATA OF THE BOUNDARY GEOMETRY ########\n" \
	          "#\n"\
	  );
  fprintf(output, "nv      ne\n");
}

void u2d_Umesh2dOutput::printverelem(FILE *output) {
  if(vertexs > 9) fprintf(output, "%i", vertexs);
  else fprintf(output, " %i", vertexs);
  if(elements > 9) fprintf(output, "      %i\n", elements);
  else fprintf(output, "      %i\n", elements);
}

inline void u2d_Umesh2dOutput::printish(FILE *output) {
  /*  fprintf(output, "   idata   sdata   hdata\n"); */
  fprintf(output, "       i       s       h\n");
}

inline void u2d_Umesh2dOutput::printenbvev(FILE *output) {
  fprintf(output, "       e       n      bv      ev\n");
}

void u2d_Umesh2dOutput::printprimg(FILE *output, Primitive *curprim) {
    if(strlen(curprim->primname) > 0){
      printbegin(output, curprim->element, curprim->primname);
    }
    else{
      printbegin(output, curprim->element);
    }
  switch(curprim->type) {
    case PR_LINE:
      fprintf(output, "line\n");
      fprintf(output, "%i %.2f %.2f %.2f %.2f\n", ((Line *)curprim)->spline, ((Line *)curprim)->x1, ((Line *)curprim)->y1, ((Line *)curprim)->x2, ((Line *)curprim)->y2);
      printend(output);
      break;
    case PR_CIRCLE:
      fprintf(output, "circle\n");
      fprintf(output, "%i %.2f %.2f %.2f %.2f %.2f\n", ((Circle *)curprim)->spline, ((Circle *)curprim)->centerx, ((Circle *)curprim)->centery, ((Circle *)curprim)->radius, ((Circle *)curprim)->startarc, ((Circle *)curprim)->endarc);
      printend(output);
      break;
    case PR_ELLIPSE:
      fprintf(output, "ellipse\n");
      fprintf(output, "%i %.2f %.2f %.2f %.2f %.2f %.2f\n", ((Ellipse *)curprim)->spline, ((Ellipse *)curprim)->centerx, ((Ellipse *)curprim)->centery, ((Ellipse *)curprim)->semix, ((Ellipse *)curprim)->semiy, ((Ellipse *)curprim)->startarc, ((Ellipse *)curprim)->endarc);
      printend(output);
      break;
    case PR_DATA:
      fprintf(output, "data\n");
      fprintf(output, "%s\n", ((Data *)curprim)->filename);
      fprintf(output, "   translation\n");
      fprintf(output, "   %.2f   %.2f\n", ((Data *)curprim)->trasx,\
	      ((Data *)curprim)->trasy);
      fprintf(output, "   rotation\n");
      fprintf(output, "   %.2f\n", ((Data *)curprim)->rotation);
      printend(output);
      break;
    default:
      fprintf(stderr, "PRINTING: Unknown primitive type\n");
    }
}

void u2d_Umesh2dOutput::printprimd(FILE *output, Primitive *curprim) {
  printbegin(output, curprim->element);
  printenbvev(output);
  fprintf(output, "       %i       %i       %i       %i\n", \
	  curprim->element, curprim->curdata, \
	  curprim->beginvertex, curprim->endvertex);
  printish(output);
  if(curprim->data != NULL) {
    for(int j = 0; j < curprim->curdata; j++) {
      fprintf(output, "       %.0f    %.2f    %.2f\n", curprim->getdata(j, 0), curprim->getdata(j, 1), curprim->getdata(j, 2));
    }
  }
  printend(output);
}

void u2d_Umesh2dOutput::printbegin(FILE *output, int id) {
  fprintf(output, "BEGIN %i\n", id);
}

void u2d_Umesh2dOutput::printbegin(FILE *output, int id, char *name) {
  fprintf(output, "BEGIN %i (%s)\n", id, name);
}

inline void u2d_Umesh2dOutput::printend(FILE *output) {
  fprintf(output, "END\n");
}

/* add firstprimitive as an argument to printptiminfo and printprimdata */
void u2d_Umesh2dOutput::printpriminfo(FILE *output) {
  Primitive *currentprimitive;
  currentprimitive = firstprimitive;
  for(int i = 1; i <= elements; i++) {
    printprimg(output, currentprimitive);
    if(currentprimitive->nextprimitive != NULL) {
      printhash(output);
      currentprimitive = currentprimitive->nextprimitive;
    }
    else if (currentprimitive->nextprimitive == NULL && i != elements) {
      fprintf(stderr, "printpriminfo: WTF!? How we are here?, elements say we still have" \
                      "primitive to print but currentprimitive->nextprimitive" \
                      "pointer is NULL\n");
    }
  }
}

void u2d_Umesh2dOutput::printprimdata(FILE *output) {
  Primitive *currentprimitive;
  currentprimitive = firstprimitive;
  for(int i = 1; i <= elements; i++) {
    printprimd(output, currentprimitive);
    if(currentprimitive->nextprimitive != NULL) {
      printhash(output);
      currentprimitive = currentprimitive->nextprimitive;
    }
    else if (currentprimitive->nextprimitive == NULL && i != elements) {
      fprintf(stderr, "printprimdata: WTF!? How we are here?, elements say we still" \
                      "have primitive to print but currentprimitive->nextprimitive" \
                      "pointer is NULL\n");
    }
  }
}
