/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2006 Gabriele Dini Ciacci
    Copyright (C) 2007 Gabriele Dini Ciacci

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <u2d_load_plotmtv.h>
#include <u2d_GLmesh.h>
#include <strings.h>
#include <string.h>
#include <string>
#include <GL/gl.h>
#include <stdlib.h>
#include <limits.h>


#define MAX_LINE_LENGTH 512
#define START_ARRAY_SIZE 10
#define ARRAY_SIZE_MULT 2

using namespace std;

u2d_Load_PlotMtv::u2d_Load_PlotMtv(const char *filename){
  filename_expanded = NULL;
  file = NULL;
  try {
    set_filename(filename);
    open_plotmtv_file();
    read_plotmtv_file();
  }
  catch (string message){
    fprintf(stderr, "%s", message.c_str());
    // Change this to a special type exception when we impelement
    // exception at higher levels
//    throw "Cannot load plotmtv file\n";
  }
}

u2d_Load_PlotMtv::~u2d_Load_PlotMtv(){

}

int u2d_Load_PlotMtv::set_filename(const char *filename){
  int path_max;
  /* Expand filename */
#ifdef PATH_MAX
  path_max = PATH_MAX;
#else
  path_max = pathconf (filename, _PC_PATH_MAX);
  if (path_max <= 0)
    path_max = 4096;
#endif
  filename_expanded = (char *)malloc(path_max * sizeof(char));
  if(filename_expanded == NULL){
    throw string("Cannot allocate memory for pathname expansion\n");
  }
  memset(filename_expanded, 0, path_max);
  realpath(filename, filename_expanded);
  return 0;
}

void u2d_Load_PlotMtv::open_plotmtv_file(){
  file = fopen(filename_expanded, "r");
  if(file == NULL){
    throw string("File opening failed\n");
  }
}

/**
 * Parse a mtv style file. This function starts by looking for a
 * line contatining one of the control characters, ignoring all but a
 * $, when one is found it tries to extract the header type from it by
 * calling parse_header(). On success a variable checking if we are
 * inside a valid header is set. From now on subsequent loops will not
 * ignore anymore the other control chars and parse them acordingly.
 * This function can parse more than one data type in the same file,
 * given that parse_header() know how to parse it and given that you
 * add what to do with the acquired data to the function.
 *
 * The loop more or less will:
 * - Call parse_header() if it finds another $ symbol before
 *   any number or % or # or . . In this case it will just switch to
 *   the new data type if it is a know data type, or unset the
 *   variable that keeps track if we are in a valid header.
 * - Call parse_option() if a % is found before any other matching
 *   symbol.
 * - Call parse_foo() where foo is the appropriate data_type if the
 *   first non-white char is a digit or a +-.
 *
 * On exit from the loop if we was scanning a valid data type the
 * display_set is pushed in the vector holding them.
 *
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Load_PlotMtv::read_plotmtv_file(){
    char buffer[MAX_LINE_LENGTH];
    memset(buffer, 0, MAX_LINE_LENGTH * sizeof(char));
    bool valid_datatype_found = false;
    int data_type = UNDEFINED;
    char *cur_pos;
    unsigned int cur_line = 0;
    u2d_Display_Set *m_display_set;

    printf("--== Plotmtv datafile acquisition started ==--\n");
    while(fgets(buffer, MAX_LINE_LENGTH, file) != NULL) {
        cur_line++;
//        printf("Parsing line %u\n", cur_line);
        cur_pos = NULL;
        cur_pos = strpbrk(buffer, "$#%1234567890.-+");
        if(cur_pos == NULL) {
//            printf(" ");
            /* Was a blank line or a line with letters.
             * We do not care
             */
        }
        else if(valid_datatype_found == false) {
            data_type = parse_header(cur_pos, strlen(cur_pos));
            if(data_type != UNDEFINED) {
                valid_datatype_found = true;
                m_display_set = new u2d_Display_Set;
                select_appropriate_GLtype(data_type, m_display_set);
            }
            else{
                printf("Undefined header found at line %i, skipping to next datatype\n", cur_line);
            }
        }
        else {
            switch(*cur_pos){
            case '$':{
#ifdef DEBUGGING
                printf("$");
#endif
                data_type = parse_header(buffer, strlen(buffer));
                if(data_type == UNDEFINED){
                    /* We switch to the new data type or
                     * if we do not know what it was,
                     * we just ignore everything until we
                     * can find again some data we are able
                     * to read
                     */
                    valid_datatype_found = false;
                    printf("Undefined header found at line %u, skipping to next datatype\n", cur_line);
                    /* We push the display set we are using, if we will find some
                     * more data that needs a display set the outer function will
                     * care of creating another
                     */
                    display_set.push_back(*m_display_set);
                    /* vertex_array pointer inside the display set should get
                     * copied over and so should not go lost
                     */
                    delete m_display_set;
                }
                else {
                    /* We found a different set of data.
                     * We generate a new display_set to hold it
                     */
                    display_set.push_back(*m_display_set);
                    /* vertex_array pointer inside the display set should get
                     * copied over and so should not go lost
                     */
                    delete m_display_set;
                    m_display_set = new u2d_Display_Set;
                    select_appropriate_GLtype(data_type, m_display_set);
                }
                break;
            }
            case '%':{
#ifdef DEBUGGING
                printf("%%");
#endif
                /* Parse the option */
                int val;
                val = parse_option(cur_pos, m_display_set);
                if(val != 0) {
                    fprintf(stderr, "Error parsing option at line %u, ignoring it\n", cur_line);
                }
                break;
            }
            case '#':{
#ifdef DEBUGGING
                printf("#");
#endif
                /* We just ignore because this is a comment */
                break;
            }
            default:{
#ifdef DEBUGGING
                printf(".");
#endif
                if(data_type == CURVE2D) {
                    int val;
                    val = parse_CURVE2D(cur_pos, m_display_set);
                    if(val != 0) {
                        fprintf(stderr, "Error parsing line %i, while reading CURVE2D datatype\n", cur_line);
                    }
                }
                else {
                    fprintf(stderr, "Tring to parse an unknown data type at line %u, this is a bug\n", cur_line);
                }
                break;
            }
            }
        }
    }
    printf("\n--== Plotmtv datafile acqusition done ==--\n");
    /* We finished reading the file while we was scanning a valid
     * data type so we push the display_set since it would go
     * lost otherwise. Other display_set s were already pushed
     * when we encountered a new data type (so it's like push old
     * and create new).
     */
    if(valid_datatype_found == true) {
        display_set.push_back(*m_display_set);
        /* vertex_array pointer inside the display set should get
         * copied over and so should not go lost
         */
        delete m_display_set;
    }
}

int u2d_Load_PlotMtv::parse_CURVE2D(char *cur_pos,\
                                    u2d_Display_Set *m_display_set){
    /* A digit or a +- */
    float a,b,c;
    int val;
    /* NOTE: c is aquired but ignored totally */
    val = sscanf(cur_pos, "%f %f %f\n", &a, &b, &c);
    if(val != 3) {
        return -1;
    }
    else {
        m_display_set->Push_Coord(a, b);
    }
    return 0;
}

/* returns the data type */
int u2d_Load_PlotMtv::parse_header(char *header, int length){
    char *cur_pos;
    if(length <= 1){
        return UNDEFINED; /* string is no length or has just the \0 */
    }

    cur_pos = strstr(header, "data");
    if(cur_pos == NULL) {
        cur_pos = strstr(header, "DATA");
    }
    if(cur_pos != NULL){
        cur_pos = index(cur_pos, '=');
        if(cur_pos == NULL) {
            fprintf(stderr, "No equal sign found in header string\n");
        }
        cur_pos = skip_white_spaces(cur_pos + 1, strlen(cur_pos + 1));
        if(cur_pos == NULL) {
            fprintf(stderr, "No value found after equal sign in header string\n");
        }
        size_t line_length;
        const char reject[] = "\n";
        line_length = strcspn(cur_pos, reject);
        for(int j = 0; j < KNOWN_FOMARTS; j++){
            if(strncasecmp(cur_pos, format_pairs[j].label, line_length) == 0){
                printf("\nFound header for \"%s\". Parsing...\n", format_pairs[j].label);
                return format_pairs[j].value;
            }
        }
        fprintf(stderr, "Specified format \"%s\" does not match any know format\n", cur_pos);
        return UNDEFINED;
    }
    printf("Unknown header type, aborting\n");
    /* exit for now */
    return UNDEFINED;
}

/*
 * Gets the first occurence to a non white or tabular char.
 * @return a char* to the first occurence of the first non while
 * or tab char.
 * @return NULL if we have reached the end of the line without
 * finding any nonwhite char
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
char *u2d_Load_PlotMtv::skip_white_spaces(char *str, int length){
    size_t start_idx;
    const char accept[] = " \t";
    start_idx = strspn(str, accept);
/* OLD WAY
  while((str[starting_index] == ' ' || str[starting_index] == '\t' ) \\
	&& starting_index <= length){
    starting_index++;
  }
  return starting_index;
*/
    if(start_idx >= length) {
        return NULL;
    }
/*    printf("start_idx is %u\n", start_idx); */
    return str + start_idx;
}

int u2d_Load_PlotMtv::parse_option(char *cur_pos,\
                                    u2d_Display_Set *m_display_set){
    return 0;
}

void u2d_Load_PlotMtv::select_appropriate_GLtype(int data_type,\
                                                 u2d_Display_Set *m_display_set){
    if(data_type == CURVE2D) {
        m_display_set->vertex_type = GL_LINES;
        m_display_set->Set_Coord_Per_Vertex(2);
    }
}
