/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <string>
#include <fstream>
#include <iostream>
#include <cstdio>

#include "u2d_mesh.h"
#include "u2d_load_umesh2d.h"
#include "exceptions.h"

using namespace std;

u2d_Load_Umesh2d::u2d_Load_Umesh2d(string path, string probname,
        bool reverse /* = false */) {
    /* Initializa the vars */
    mesh = new u2d_Mesh::u2d_Mesh();
    d_elem_n = 0;
    b_elem_n = 0;
    d_nodes_n = 0;
    b_nodes_n = 0;
    spatial_dim = 0;
    /* Be tollerant if the path already ends with a / or does not */
    size_t last_slash_index = path.find_last_of("/");
    if (last_slash_index == path.npos ||
            last_slash_index != (path.length() - 1)) {
        cout << "No trailing slash found, adding" << endl;
        path = path + "/";
    }

    /* Assemble the names */
    grid_filename = path + "grid." + probname;
    nodes_filename = path + "nodes." + probname;

    cout << " grid: " << grid_filename << endl;
    cout << " nodes: " << nodes_filename << endl;

    try {
        open_umesh2d_file();
        parse_umesh2d_file();
        close_umesh2d_file();
        mesh->setBoundaryNodeReversal(reverse);
        mesh->getReady();
    } catch (u2d_Parse_Error_Exception &) {
        throw;
    }


}

u2d_Load_Umesh2d::~u2d_Load_Umesh2d() {
    close_umesh2d_file();
}

void u2d_Load_Umesh2d::open_umesh2d_file() {
    grid_file.open(grid_filename.c_str(), ifstream::out);
    nodes_file.open(nodes_filename.c_str(), ifstream::out);
}

void u2d_Load_Umesh2d::close_umesh2d_file() {
    if (grid_file.is_open()) {
        grid_file.close();
    }
    if (nodes_file.is_open()) {
        nodes_file.close();
    }
}

void u2d_Load_Umesh2d::parse_umesh2d_file() {
    parse_grid_size();
    parse_nodes_size();
    /* Now we can resize the mesh to correct size */
    mesh->resize(spatial_dim, d_nodes_n, b_nodes_n, d_elem_n, b_elem_n);
    parse_grid();
    parse_nodes();
}

void u2d_Load_Umesh2d::parse_grid_size()
throw (u2d_Parse_Error_Exception &) {
    string line;
    bool found = false;
    bool goodstream = true;
    cout << "--== Reading grid size  ==--" << endl;
    while (!found && goodstream) {
        goodstream = getline(grid_file, line).good();
        string::size_type idx = line.find_first_not_of(" \t\n\v");
        if (idx != line.npos && line[idx] == '#') {
            /* Line was starting with a # */
        } else if (idx == line.npos) {
            /* Line was empty */
        } else {
            /* Now we found the first good line, so we scan it and exit */
            sscanf(line.c_str(), "%d %d", &d_elem_n, &b_elem_n);
            found = true;
        }
    }
    if (found == false) {
        throw new u2d_Parse_Error_Exception("Cannot find grid sizes");
    }
    cout << "Domain elements: " << d_elem_n << endl;
    cout << "Boundary elements: " << b_elem_n << endl;
    cout << "--== Done ==--" << endl;
}

void u2d_Load_Umesh2d::parse_nodes_size()
throw (u2d_Parse_Error_Exception &) {
    string line;
    bool found = false;
    cout << "--== Reading nodes size  ==--" << endl;
    while (!found) {
        getline(nodes_file, line);
        string::size_type idx = line.find_first_not_of(" \t\n\v");
        if (idx != line.npos && line[idx] == '#') {
            /* Line was starting with a # */
        } else if (idx == line.npos) {
            /* Line was empty */
        } else {
            /* Now we found the first good line, so we scan it and exit */
            sscanf(line.c_str(), "%d %d %d", &spatial_dim, &d_nodes_n, &b_nodes_n);
            found = true;
        }
    }
    if (spatial_dim != 2) {
        throw new u2d_Parse_Error_Exception("Spatial dimensions are not 2, unsupported");
    } else if (found == false) {
        throw new u2d_Parse_Error_Exception("Cannot find nodes sizes");
    }
    cout << "Domain nodes: " << d_nodes_n << endl;
    cout << "Boundary nodes: " << b_nodes_n << endl;
    cout << "--== Done ==--" << endl;
}

void u2d_Load_Umesh2d::parse_grid() {
    string line;
    cout << "--== Reading grid elements  ==--" << endl;
    /* DOMAIN ELEMENTS */
    for (int i = 0; i < d_elem_n; i++) {
        /* Information is given in triplets (tree lines),
         *  so we analyze each separatelly */
        for (int j = 0; j < 3; j++) {
            getline(grid_file, line);
            string::size_type idx = line.find_first_not_of(" \t\n\v");
            if (idx != line.npos && line[idx] == '#') {
                /* Line was empty or starting with a # 
                 * We decrement the counter cause we did not got an element line
                 * it was just a comment line */
                j--;
            } else {
                if (j == 0) {
                    // just jump first line, we are not interested
                    // sscanf(line.c_str(), "%d %d", &dummy, &dummy);
                    //cout << line << endl;
                } else if (j == 1) {
                    int a, b, c;
                    sscanf(line.c_str(), "%d %d %d", &a, &b, &c);
                    mesh->jjd(0, i) = a;
                    mesh->jjd(1, i) = b;
                    mesh->jjd(2, i) = c;
                    //cout << line << "  read: " << a << " " << b << " " << c << endl;
                    //cout << "saved: " << mesh->jjd(0, i) << " " << mesh->jjd(1, i) << " " << mesh->jjd(2, i) << endl;
                } else if (j == 2) {
                    int a, b, c;
                    sscanf(line.c_str(), "%d %d %d", &a, &b, &c);
                    mesh->neigh(0, i) = a;
                    mesh->neigh(1, i) = b;
                    mesh->neigh(2, i) = c;
                    //cout << line << "  read: " << a << " " << b << " " << c << endl;
                    //cout << "saved: " << mesh->neigh(0, i) << " " << mesh->neigh(1, i) << " " << mesh->neigh(2, i) << endl;
                }
            }
        }
    }
    /* BOUNDARY ELEMENTS */
    for (int i = 0; i < b_elem_n; i++) {
        /* Information is given in triplets (tree lines),
         *  so we analyze each separatelly */
        for (int j = 0; j < 3; j++) {
            getline(grid_file, line);
            string::size_type idx = line.find_first_not_of(" \t\n\v");
            if (idx != line.npos && line[idx] == '#') {
                /* Line was empty or starting with a #
                 * We decrement the counter cause we did not got an element line
                 * it was just a comment line */
                j--;
            } else {
                if (j == 0) {
                    int a, b, c;
                    sscanf(line.c_str(), "%d %d %d", &a, &b, &c);
                    mesh->neigh_s(i) = a;
                    /* b contains the type of boundary element,
                     * we are not interested in it, we discart it! */
                    mesh->b_id(i) = c;
                    //cout << line << " read: " << a << " " << c << endl;
                    //cout << "saved: " << mesh->neigh_s(i) << " " << mesh->b_id(i) << endl;
                } else if (j == 1) {
                    int a, b;
                    sscanf(line.c_str(), "%d %d", &a, &b);
                    mesh->jjb_s(0, i) = a;
                    mesh->jjb_s(1, i) = b;
                    //cout << line << "  read: " << a << " " << b << " " << endl;
                    //cout << "saved: " << mesh->jjb_s(0, i) << " " << mesh->jjb_s(1, i) << endl;
                } else if (j == 2) {
                    // just jump first line, we are not interested
                    //sscanf(line.c_str(), "%d %d %d", &a, &b, &c);
                    //cout << line << endl;
                }
            }
        }
    }
    cout << "--== Done ==--" << endl;
}

void u2d_Load_Umesh2d::parse_nodes() {
    string line;
    cout << "--== Reading nodes  ==--" << endl;
    /* DOMAIN NODES */
    for (int i = 0; i < d_nodes_n; i++) {
        /* Information is given in couples (two lines),
         *  so we analyze each separatelly */
        for (int j = 0; j < 2; j++) {
            getline(nodes_file, line);
            string::size_type idx = line.find_first_not_of(" \t\n\v");
            if (idx != line.npos && line[idx] == '#') {
                /* Line was empty or starting with a #
                 * We decrement the counter cause we did not got an element line
                 * it was just a comment line */
                j--;
            } else {
                if (j == 0) {
                    // just jump first line, we are not interested
                    // sscanf(line.c_str(), "%d %d", &dummy, &dummy);
                    //cout << line << endl;
                } else if (j == 1) {
                    double a, b;
                    sscanf(line.c_str(), "%lE %lE", &a, &b);
                    mesh->rr(0, i) = a;
                    mesh->rr(1, i) = b;
                    //cout << line << "  read: " << a << " " << b << " " << endl;
                    //cout << "saved: " << mesh->rr(0, i) << " " << mesh->rr(1, i) << endl;
                }
            }
        }
    }
    /* BOUNDARY NODES */
    for (int i = 0; i < b_nodes_n; i++) {
        /* Information is given in single lines */
        getline(nodes_file, line);
        string::size_type idx = line.find_first_not_of(" \t\n\v");
        if (idx != line.npos && line[idx] == '#') {
            /* Line was empty or starting with a #
             * We decrement the counter cause we did not got an element line
             * it was just a comment line */
            i--;
        } else {
            int a, b, c;
            sscanf(line.c_str(), "%d %d %d", &a, &b, &c);
            mesh->jjb_sd(i) = b;
            //cout << line << "  read: " << a << " " << endl;
            //cout << "saved: " << mesh->jjb_sd(i) << endl;
        }
    }
    cout << "--== Done ==--" << endl;
}

u2d_Mesh *u2d_Load_Umesh2d::getMesh() {
    return mesh;
}

void u2d_Load_Umesh2d::destoryMesh(u2d_Mesh *_mesh) {
    delete _mesh;
}
