/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on April 16, 2010, 4:31 PM
 */

#include "boost/numeric/mtl/mtl.hpp"
#include "u2d_load_vectorfield.h"

u2d_LoadVectorField::u2d_LoadVectorField(string _filename, int _spatial_dim, bool _lineonesize, size_t _suggested_size) :
u2d_LoadValues(_filename, _lineonesize, _suggested_size),
read_vector(0, 0),
vector_size(0),
spatial_dim(_spatial_dim) {
    parseValuesFile();
}

u2d_LoadVectorField::~u2d_LoadVectorField() {
}

dense2D<double> u2d_LoadVectorField::getField() {
    return read_vector;
}

void u2d_LoadVectorField::parseValuesFile()
throw (u2d_Parse_Error_Exception &) {
    if (lineonesize) {
        parseSize();
    }
    parseElements();
}
/**
 * Parse the first line of the file conting the size.
 * The size should be in the form of an integer number on the first valid
 * line (not starting with an #) of the file. i.e.:
 * # A scalar filed
 * # size
 * 345
 * # The data evalueted on nodes
 * 455.76 45.87
 * 45.78  56.867
 * ....  ....
 * ....  ....
 * # the above was last element, that is 345th
 * # any line of comment can be here or empty line
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_LoadVectorField::parseSize()
throw (u2d_Parse_Error_Exception &) {
    string line;
    bool found = false;
    cout << "--== Reading vector size  ==--" << endl;
    while (!found && getline(filestream, line).good()) {
        string::size_type idx = line.find_first_not_of(" \t\n\v");
        if (idx != line.npos && line[idx] == '#') {
            /* Line was starting with a # */
        } else if (idx == line.npos) {
            /* Line was empty */
        } else {
            /* Now we found the first good line, so we scan it and exit */
            sscanf(line.c_str(), "%d", &vector_size);
            found = true;
        }
    }
    if (found == false) {
        throw new u2d_Parse_Error_Exception("Cannot find vector size");
    }
    else if (suggested_size > 0 && vector_size != suggested_size) {
        string errormsg = "Parsed vector size is: " + vector_size;
        string errormsg2 = " but suggested size was: " + suggested_size;
        throw new u2d_Parse_Error_Exception(errormsg + errormsg2);
    }
    cout << "Vector elements: " << vector_size << endl;
    if(suggested_size != 0) {
        cout << "Vector elements (suggested): " << suggested_size << endl;
    }
    cout << "--== Done ==--" << endl;
}


void u2d_LoadVectorField::parseElements()
throw (u2d_Parse_Error_Exception &) {
    string line;
    std::vector<double> valuesx;
    std::vector<double> valuesy;
    std::vector<double> valuesz;
    if(vector_size > 0) {
        valuesx.reserve(vector_size);
        valuesy.reserve(vector_size);
        valuesz.reserve(vector_size);
    }
    cout << "--== Reading vector elements  ==--" << endl;
    while (getline(filestream, line).good()) {
        double x = 0.0;
        double y = 0.0;
        double z = 0.0;
        string::size_type idx = line.find_first_not_of(" \t\n\v");
        if (idx != line.npos && line[idx] == '#') {
            /* Line was starting with a # */
        } else if (idx == line.npos) {
            /* Line was empty */
        } else {
            /* Now we found the first good line, so we scan it and exit */
            if (spatial_dim == 2) {
                sscanf(line.c_str(), "%lE %lE", &x, &y);
                valuesx.push_back(x);
                valuesy.push_back(y);
            } else if (spatial_dim == 3) {
                sscanf(line.c_str(), "%lE %lE %lE", &x, &y, &z);
                valuesx.push_back(x);
                valuesy.push_back(y);
                valuesz.push_back(z);
            } else {
                string errormsg = "Unsupported spatial dimension: " + spatial_dim;
                throw new u2d_Parse_Error_Exception(errormsg);
            }
        }
    }
    // TODO Checking if the exit was for error or oef and throw if needed
    if (suggested_size > 0 && valuesx.size() != suggested_size) {
        string errormsg = "Parsed element number is: " + valuesx.size();
        string errormsg2 = " but suggested size was: " + suggested_size;
        throw new u2d_Parse_Error_Exception(errormsg + errormsg2);
    } else if(vector_size > 0 && vector_size != valuesx.size()) {
        string errormsg = "Parsed element number is: " + valuesx.size();
        string errormsg2 = " but size specified in file was: " + vector_size;
        throw new u2d_Parse_Error_Exception(errormsg + errormsg2);
    }
    /* Convert the vector to mtl vector */
    read_vector.change_dim(spatial_dim, valuesx.size());
    size_t index = 0;
    if (spatial_dim == 2) {
        std::vector<double>::iterator ity = valuesy.begin();
        for (std::vector<double>::iterator itx = valuesx.begin(); itx != valuesx.end(); ++itx) {
            read_vector(0, index) = *itx;
            read_vector(1, index) = *ity;
            ++ity;
            index++;
        }
    } else if (spatial_dim == 3) {
        std::vector<double>::iterator ity = valuesy.begin();
        std::vector<double>::iterator itz = valuesz.begin();
        for (std::vector<double>::iterator itx = valuesx.begin(); itx != valuesx.end(); ++itx) {
            read_vector(0, index) = *itx;
            read_vector(1, index) = *ity;
            read_vector(2, index) = *itz;
            ++ity;
            ++itz;
            index++;
        }
    }
    cout << "--== Done ==--" << endl;
}
