/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on February 22, 2010, 12:34 PM
 */

#include <math.h>

#include "u2d_Constants.h"
#include "boost/numeric/mtl/mtl.hpp"

using namespace mtl;
using namespace std;

u2d_Constants::u2d_Constants() {
}

u2d_Constants::u2d_Constants(const u2d_Constants& orig) {
}

u2d_Constants::~u2d_Constants() {
}

/**
 * Computes the Thermal Conductivity vector k.
 * K has the same number of columns as rr and is computed by the formula:
 *   \f[ k(r) = 1 + |r|^2 \f]
 * hence k > 0.
 * @param <rr> are the nodes where to evaluate the function k(r). The exptected
 * form is:
 *         node 1  node 2 ...  node 3  node n
 *  rr = | 45.756  34.43  ...  0.3459  0.3428 |  coord x
 *       | 765.32  45.32  ...  4.6426  0.6583 |  coord y
 *
 */
dense_vector<double> u2d_Constants::thermalConductivity(dense2D<double> rr) {
    size_t nodes_n = num_cols(rr);
    dense_vector<double> k(nodes_n, 0.0);
    for (size_t i = 0; i < nodes_n; i++) {
        k[i] = 1 + pow(rr[0][i], 2) + pow(rr[1][i], 2);
    }
    return k;
}

dense2D<double> u2d_Constants::gradThermalConductivity(dense2D<double> rr) {
    size_t spatial_dim = num_rows(rr);
    size_t nodes_n = num_cols(rr);
    dense2D<double> gradk(spatial_dim, nodes_n);
    for (size_t i = 0; i < nodes_n; i++) {
        gradk[0][i] = 2 * rr[0][i];
        gradk[1][i] = 2 * rr[1][i];
    }
    return gradk;
}

/**
 * Computes v, the vector field of speed of the rigid rotation.
 *
 * @param rr are the nodes where to evaluate the function k(r). The exptected
 * form is:
 *         node 1  node 2 ...  node 3  node n
 *  rr = | 45.756  34.43  ...  0.3459  0.3428 |  coord x
 *       | 765.32  45.32  ...  4.6426  0.6583 |  coord y
 */
dense2D<double> u2d_Constants::advection(dense2D<double> rr) {
    size_t spatial_dim = num_rows(rr);
    size_t nodes_n = num_cols(rr);
    dense2D<double> v(spatial_dim, nodes_n);

    double omega = 1.0;
    for (size_t i = 0; i < nodes_n; i++) {
        v[0][i] = -rr(1, i) * omega;
        v[1][i] = rr(0, i) * omega;
    }
    return v;
}

dense_vector<double> u2d_Constants::pythagoras(dense2D<double> rr) {
    size_t nodes_n = num_cols(rr);
    dense_vector<double> u(nodes_n, 0.0);
    for (size_t i = 0; i < nodes_n; i++) {
        u[i] = pow(rr[0][i], 2) + pow(rr[1][i], 2);
    }
    return u;
}

dense2D<double> u2d_Constants::gradPythagoras(dense2D<double> rr) {
    size_t spatial_dim = num_rows(rr);
    size_t nodes_n = num_cols(rr);
    dense2D<double> gradu(spatial_dim, nodes_n);
    for (size_t i = 0; i < nodes_n; i++) {
        gradu[0][i] = 2 * rr(0, i);
        gradu[1][i] = 2 * rr(1, i);
    }
    return gradu;
}

dense_vector<double> u2d_Constants::laplacianSolution(dense2D<double> rr) {
    size_t nodes_n = num_cols(rr);
    dense_vector<double> lsol(nodes_n, 4.0);
    return lsol;
}

dense_vector<double> u2d_Constants::source(dense2D<double> rr, double gamma) {
    size_t spatial_dim = num_rows(rr);
    size_t nodes_n = num_cols(rr);
    dense_vector<double> k(nodes_n, 0.0);
    dense2D<double> gradk(spatial_dim, nodes_n);
    dense2D<double> gradu(spatial_dim, nodes_n);
    dense2D<double> v(spatial_dim, nodes_n);
    dense_vector<double> sol(nodes_n, 0.0);
    dense_vector<double> lsol(nodes_n, 0.0);
    dense_vector<double> f(nodes_n, 0.0);

    k = u2d_Constants::thermalConductivity(rr);
    gradk = u2d_Constants::gradThermalConductivity(rr);
    gradu = u2d_Constants::gradPythagoras(rr);
    v = u2d_Constants::advection(rr);
    lsol = laplacianSolution(rr);
    sol = pythagoras(rr);

    for (int i = 0; i < nodes_n; i++) {
        f[i] = -k[i] * lsol[i] \
                - gradk(0, i) * gradu(0, i) - gradk(1, i) * gradu(1, i) \
                + v(0, i) * gradu(0, i) + v(1, i) * gradu(1, i) \
                + gamma * sol(i);

    }
    return f;

}