/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Created on January 20, 2010, 7:22 PM
 */

#include <algorithm> // for find()

#include "u2d_DirichletCondition.h"
#include "u2d_mesh.h"
#include "boost/numeric/mtl/mtl.hpp"

using namespace mtl;
using namespace std;

u2d_DirichletCondition::u2d_DirichletCondition() :
dirichlet_nodes_n(0),
dirichlet_nodes(0) {
}

u2d_DirichletCondition::u2d_DirichletCondition(const u2d_DirichletCondition& orig) {
}

u2d_DirichletCondition::~u2d_DirichletCondition() {
}

/**
 * Count all nodes of the boundary elements which belong to
 * the sides where a Dirichlet boundary condition is prescribed.
 */
void u2d_DirichletCondition::countNodes(u2d_Mesh* mesh) {
    /* If the same node belongs to different sides of the boundary
     * the algorithm takes it into account only once.
     */
    dirichlet_nodes_n = 0;
    for (int i = 0; i < mesh->getBoundaryElementNumber(); i++) {
        if (isDirichlet(mesh->getElementSideId(i + 1))) {
            dirichlet_nodes_n += BOUNDARY_ELEMENT_SIZE; // because each element has BOUNDARY_ELEMENT_SIZE nodes
        }
    }
    /* Resize the vector accordingly */
    dirichlet_nodes.change_dim(dirichlet_nodes_n);

    // Lovely debug output
    //cout << dirichlet_nodes_n << endl;
}

/**
 * Generates the nodes list where to enforece Dirichlet condition.
 */
void u2d_DirichletCondition::generateNodes(u2d_Mesh* mesh) {
    /* Count the nodes and prepare for run */
    countNodes(mesh);
    /* Clean the set, just in case this is a rerun */
    dirichlet_unique_nodes.erase(dirichlet_unique_nodes.begin(), dirichlet_unique_nodes.end());
    /*  Let's go */
    int cur_node = 0;
    for (int i = 0; i < mesh->getBoundaryElementNumber(); i++) {
        int element_id = i + 1;
        int element_idx = i;
        if (isDirichlet(mesh->getElementSideId(element_id))) {
            /* Builds a mtl vector containing all nodes, even duplicates */
            dirichlet_nodes(cur_node) = mesh->jjb(0, element_idx);
            dirichlet_nodes(cur_node + 1) = mesh->jjb(1, element_idx);
            /* Builds a std::set containing only unique nodes and that is
             * already sorted */
            dirichlet_unique_nodes.insert(mesh->jjb(0, element_idx));
            dirichlet_unique_nodes.insert(mesh->jjb(1, element_idx));
            /* Increment the current node count, we processed two node ids,
             * one for each extreme. */
            cur_node += 2; // 2 because each element has 2 nodes
        }
    }
    dirichlet_unique_nodes_n = dirichlet_unique_nodes.size();
    // Lovely debug output
    //cout << dirichlet_unique_nodes_n << endl;
}

/**
 * Set the given side id as a side where to enforce
 * the DirichletCondition.
 * @return the current number of sides that have been set.
 * i.e. The total number of sides where the Dirichlet boundary condition
 * is to be enforced:
 * DiricheletCondition dc;
 * dc->setside(4) // return 1
 * dc->setside(1) // return 2
 * dc->setside(5) // return 3
 */
int u2d_DirichletCondition::setSide(int side_id) {
    dir_sides.push_back(side_id);
    return dir_sides.size();
}

/**
 * Say if a given side id is in the list of sides where to enforce
 * the Disichlet boundary condition.
 * @return Return true if the side_id is in the list and Dirichlet condtion
 * must be enforced on it. Return false otherwise.
 */
bool u2d_DirichletCondition::isDirichlet(int side_id) {
    if (find(dir_sides.begin(), dir_sides.end(), side_id) != dir_sides.end()) {
        return true;
    }
    return false;
}
