/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on January 27, 2010, 6:10 PM
 */

#include "u2d_ProblemCalculator.h"
#include "u2d_DirichletCondition.h"
#include "u2d_Problem.h"
#include "u2d_mesh.h"
#include "u2d_GaussPoints.h"
#include "u2d_IsoElement2DP13Points.h"
#include "u2d_IsoElement1DP12Points.h"
#include "boost/numeric/mtl/mtl.hpp"

using namespace std;

u2d_ProblemCalculator::u2d_ProblemCalculator(u2d_Problem *_problem, u2d_Mesh *_mesh) :
problem(_problem),
mesh(_mesh) {

}

u2d_ProblemCalculator::u2d_ProblemCalculator(const u2d_ProblemCalculator& orig) {
}

u2d_ProblemCalculator::~u2d_ProblemCalculator() {
}



/********************************************************
 *                LEFT-HAND SIDE TERMS                  *
 ********************************************************/

/**
 * Enforce the Dirichlet condition. There is NO shift currently, so
 * the condition is only homogeneous.
 * This is a *left* hand side term.
 */
void u2d_ProblemCalculator::forceDirichlet_left(u2d_DirichletCondition *cond) {

    std::set<int>::iterator it;
    int problem_size = problem->getSize();

    /* CHECK: I am not sure that in this way we are not adding terms
     * where there would be none. This happen because we have a dense matrix
     * so terms at index (i, i) are always found. The original CSR version
     * would find (and update) the term (i, i) only if it was present in the
     * sparseness profile.
     */
    for (it = cond->dirichlet_unique_nodes.begin();
            it != cond->dirichlet_unique_nodes.end();
            it++) {
        for (int j = 0; j < problem_size; j++) {
            problem->setTerm(*it - 1, j, 0);
        }
        problem->setTerm(*it - 1, *it - 1, 1);
    }

    /* More readable version of the above algo.
     * But naturally uses a a method that set does not have
     * this one is good if set get uniformed to other arrays
     * and become a mtl::dense_vector. */
    /* READABLE VERSION 
    int nodes_n = cond->dirichlet_unique_nodes.size();
    for (int i = 0; it < nodes_n; i++) {
        int cur_row = cond->dirichlet_unique_nodes.getElement(i) // this method is fantasy
        for (int j = 0, j < problem_size, j++) {
            problem->setTerm(cur_row, j + 1, 0);
        }
        problem->setTerm(cur_row, cur_row, 1);
    }
     */

}

/**
 * Calculates:
 *  alpha * < W , G >
 * Where:
 * alpha is a scalar.
 * W(x, y) is the shape function.
 * G(x, y) is the test function.
 * This is a *left* hand side term.
 *
 */
void u2d_ProblemCalculator::addAlphaW__G_left(u2d_DomainGaussPoints *dgp, double alpha) {
    /* FIX ME, THE Iso Element should be deducted from DomainGaussPoints */
    /* Create an isoElement to recover weights from it */
    u2d_IsoElement2DP13Points isoElemenet;


    for (int m = 0; m < mesh->getDomainElementNumber(); m++) {
        /* We are getting the jacobian*weight at index ,0 that is
         * for element 1! Pay attention, if we plan to not integrate on
         * all points this needs to be changed to use an auxiliary var to store
         * the current element */
        double coef = alpha * dgp->jacobian_weight(m);
        for (int i = 0; i < DOMAIN_ELEMENT_SIZE; i++) {
            int row_index = mesh->jjd(i, m) - 1; // This is *already* the index
            for (int j = 0; j < DOMAIN_ELEMENT_SIZE; j++) {
                int column_index = mesh->jjd(j, m) - 1; // This is *already* the index
                double term = 0;
                for (int gp = 0; gp < isoElemenet.gauss_points_n; gp++) {
                    term += isoElemenet.w[i][gp] * coef * isoElemenet.w[j][gp];
                }
                problem->addTerm(row_index, column_index, term);
            }

        }

    }
}

/**
 * Calculates:
 *  < DW , alpha DG >
 * Where:
 * alpha is a vector of same size a mesh->jjd (number of domain nodes). It
 * represent the scalar filed alpha evaluated at nodes.
 * DW(x, y) is the derivative of the shape function.
 * DG(x, y) is the derivative of the test function.
 * This is a *left* hand side term.
 * TODO: CHECK ME FIRST IN CASE OF PROBLEMS.
 * @param
 */
void u2d_ProblemCalculator::addDW__AlphaDG_left(u2d_DomainGaussPoints *dgp, dense_vector<double> alpha) {
    /* FIX ME, THE Iso Element should be deducted from DomainGaussPoints */
    /* Create an isoElement to recover weights from it */
    u2d_IsoElement2DP13Points iso_elem;

    for (int m = 0; m < mesh->getDomainElementNumber(); m++) {
        /* Get the vector of nodes ID for current element m */
        dense_vector<double> nodes_id;
        nodes_id = mesh->getDomainElementNodesId(m);
        /* Evaluate (retrive) alpha at nodes of each element m */
        dense_vector<double> alpha_val(DOMAIN_ELEMENT_SIZE, 0.0);
        for (int i = 0; i < DOMAIN_ELEMENT_SIZE; i++) {
            alpha_val(i) = alpha(nodes_id(i) - 1);
        }
        for (int i = 0; i < DOMAIN_ELEMENT_SIZE; i++) {
            int row_index = mesh->jjd(i, m) - 1; // This is *already* the index
            for (int j = 0; j < DOMAIN_ELEMENT_SIZE; j++) {
                int column_index = mesh->jjd(j, m) - 1; // This is *already* the index
                double term = 0;
                // WARNING: 2D dependent code
                term += dgp->dw_x(i, m) * dgp->dw_x(j, m);
                term += dgp->dw_y(i, m) * dgp->dw_y(j, m);
                double coef = 0;
                for (int gp = 0; gp < iso_elem.gauss_points_n; gp++) {
                    coef += dot(alpha_val, iso_elem.w[mtl::irange()][gp]) * dgp->jacobian_weight(m);
                }
                problem->addTerm(row_index, column_index, coef * term);
            }

        }

    }
}

/**
 * Calculates:
 *  < W , alpha .dot. DG >
 * Where:
 * alpha is a matrix of size number of spatial dimensions and mesh->jjd
 * (number of domain nodes).
 * It represents the vector field alpha evaluated at nodes.
 * Example, domain with 3000 nodes in 2D, it's 2x3000.
 * W(x, y) is the shape function.
 * DG(x, y) is the derivative of the test function.
 * This is a *left* hand side term.
 */
void u2d_ProblemCalculator::addW__AlphaDG_left(u2d_DomainGaussPoints *dgp, dense2D<double> alpha) {
    /* FIX ME, THE Iso Element should be deducted from DomainGaussPoints */
    /* Create an isoElement to recover weights from it */
    u2d_IsoElement2DP13Points iso_elem;

    for (int m = 0; m < mesh->getDomainElementNumber(); m++) {
        /* Get the vector of nodes ID for current element m */
        dense_vector<double> nodes_id;
        nodes_id = mesh->getDomainElementNodesId(m);
        /* Evaluate (retrive) alpha at nodes of each element m.
         * This generates a matrix with components of alpha in each spatial
         * dimension in column, and a number of column equal to the number
         * of nodes that form an element.
         * Example in 2D for triangular elements:
         *                node 1  node 2   node 3
         *  alpha_val = | 3.545   4.657    1.656 |  coord x
         *              | 0.453   87.34    0.466 |  coord y
         *
         *  */
        dense2D<double> alpha_val(mesh->getSpatialDimension(), DOMAIN_ELEMENT_SIZE);
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            for (int i = 0; i < DOMAIN_ELEMENT_SIZE; i++) {
                alpha_val(j, i) = alpha(j, nodes_id(i) - 1);
            }
        }

        for (int i = 0; i < DOMAIN_ELEMENT_SIZE; i++) {
            int row_index = mesh->jjd(i, m) - 1; // This is *already* the index
            for (int j = 0; j < DOMAIN_ELEMENT_SIZE; j++) {
                int column_index = mesh->jjd(j, m) - 1; // This is *already* the index
                double term = 0;

                for (int gp = 0; gp < iso_elem.gauss_points_n; gp++) {
                    dense_vector<double> gl(mesh->getSpatialDimension(), 0.0);
                    gl = alpha_val * iso_elem.w[mtl::irange()][gp];
                    gl *= dgp->jacobian_weight(m);
                    /* OLD ALGORITHM TO CALCULATE gl
                     * for (int sdim = 0; sdim < mesh->getSpatialDimension(); sdim++) {
                     *    gl(k) = dot(alpha_val(sdim, mtl::iall), iso_elem.w(mtl::iall, gp))
                     *                        * dgp->jacobian_weight(m);
                     * }
                     */
                    dense_vector<double> dw_x_y(mesh->getSpatialDimension(), 0.0);
                    dw_x_y(0) = dgp->dw_x(j, m); // WARNING: 2D dependent code
                    dw_x_y(1) = dgp->dw_y(j, m); // WARNING: 2D dependent code
                    term = dot(gl, dw_x_y);
                    term *= iso_elem.w(i, gp);
                    /* OLD ALGORITHM TO CALCULATE term
                     * term = iso_elem.w[row_index][gp] * (gl(0) * dw_x(column_index, m) + gl(1) * dw_y(column_index, m));
                     */
                    problem->addTerm(row_index, column_index, term);
                }
            }

        }

    } // for each element we must integrate on
}

/**
 * Calculates:
 *  alpha * < W , G >_boundary on the BOUNDARY
 * Where:
 * alpha is a scalar
 * W(x, y) is the shape function.
 * G(x, y) is the test function.
 * This is a *left* hand side term.
 * This is a *boundary* term.
 */
void u2d_ProblemCalculator::addAlphaW__G_surf_left(u2d_BoundaryGaussPoints *bgp,
        u2d_NeumannCondition *cond, double alpha) {
    /* FIX ME, THE Iso Element should be deducted from BoundaryGaussPoints */
    /* Create an isoElement to recover weights from it */
    u2d_IsoElement1DP12Points iso_elem;

    int m;
    for (int index = 0; index < cond->getNuemannElementNumber(); index++) {
        m = cond->neumann_elements(index) - 1; // Index of the element, because of the -1
        /* We are getting the jacobian*weight at index 0 that is
         * for element 1! Pay attention, if we plan to not integrate on
         * all points this needs to be changed to use an auxiliary var to store
         * the current element */
        double coef = alpha * bgp->jacobian_weight(m);
        for (int i = 0; i < BOUNDARY_ELEMENT_SIZE; i++) {
            int row_index = mesh->jjb(i, m) - 1; // This is *already* the index
            for (int j = 0; j < BOUNDARY_ELEMENT_SIZE; j++) {
                int column_index = mesh->jjb(j, m) - 1; // This is *already* the index
                double term = 0;
                for (int gp = 0; gp < iso_elem.gauss_points_n; gp++) {
                    term += iso_elem.w[i][gp] * coef * iso_elem.w[j][gp];
                }
                problem->addTerm(row_index, column_index, term);
            }

        }

    }
}



/********************************************************
 *                RIGHT-HAND SIDE TERMS                 *
 ********************************************************/

/**
 * Enforce the Dirichlet condition.
 * This is a *right* hand side term.
 * @param <sol> This vector must be in the same order as the set of unique
 *  dirichlet nodes generated by u2d_DirichletCondition. It must also be
 * of the same size naturally.
 */
void u2d_ProblemCalculator::forceDirichlet_right(u2d_DirichletCondition *cond, dense_vector<double> sol) {

    std::set<int>::iterator it;
    size_t node_index = 0; // NOTE THIS IS ALREADY AN INDEX (not an ID )
    for (it = cond->dirichlet_unique_nodes.begin();
            it != cond->dirichlet_unique_nodes.end();
            it++) {
        problem->setRHSTerm(*it - 1, sol(node_index));
        ++node_index;
    }
}

/**
 * Calculates:
 * < W , f >
 * Where:
 * W(x, y) is the shape function.
 * f(x, y) is the know function (RHS).
 * This is a *right* hand side term.
 *
 */
void u2d_ProblemCalculator::addW__G_right(u2d_DomainGaussPoints *dgp, dense_vector<double> source) {
    /* FIX ME, THE Iso Element should be deducted from DomainGaussPoints */
    /* Create an isoElement to recover weights from it */
    u2d_IsoElement2DP13Points isoElement;


    for (int m = 0; m < mesh->getDomainElementNumber(); m++) {
        for (int gp = 0; gp < isoElement.gauss_points_n; gp++) {
            double coef = 0;
            /* Here we are multipling each element of source with
             * each element of w. We could have assembled the vectors first and
             * then used the dot product. Given the real simple case, this
             * proposed way is definitely faster, but notice the double for with
             * identical condition that is hadtly avoided in this case. */
            for (int i = 0; i < DOMAIN_ELEMENT_SIZE; i++) {
                coef += source(mesh->jjd(i, m) - 1) * isoElement.w(i, gp);
            }
            for (int i = 0; i < DOMAIN_ELEMENT_SIZE; i++) {
                double term = coef * isoElement.w(i, gp) * dgp->jacobian_weight(m);
                problem->addRHSTerm(mesh->jjd(i, m) - 1, term);
            }
        }
    }
}

/**
 * Calculates:
 * < W , f >_boundary on the BOUNDARY
 * Where:
 * W(x, y) is the shape function.
 * f(x, y) is the know function (RHS).
 * This is a *right* hand side term.
 * This is a *boundary* term.
 */
void u2d_ProblemCalculator::addW__G_surf_right(u2d_BoundaryGaussPoints *bgp,
        u2d_NeumannCondition *cond, dense_vector<double> sol) {
    /* FIX ME, THE Iso Element should be deducted from BoundaryGaussPoints */
    /* Create an isoElement to recover weights from it */
    u2d_IsoElement1DP12Points isoElement;

    int m;
    for (int index = 0; index < cond->getNuemannElementNumber(); index++) {
        m = cond->neumann_elements(index) - 1; // Index of the element, because of the -1
        for (int gp = 0; gp < isoElement.gauss_points_n; gp++) {
            double coef = 0;
            for (int i = 0; i < BOUNDARY_ELEMENT_SIZE; i++) {
                /* We use jjb_s here because the values in source have been generated in
                 * respect to the surface numeration of nodes. If source was numbered in
                 * respect to domain numeraton we would need to use jjb instead
                 */
                coef += sol(mesh->jjb_s(i, m) - 1) * isoElement.w(i, gp);
            }
            for (int i = 0; i < BOUNDARY_ELEMENT_SIZE; i++) {
                double term = coef * isoElement.w(i, gp) * bgp->jacobian_weight(m);
                problem->addRHSTerm(mesh->jjb(i, m) - 1, term);
            }
        }
    }
}

/**
 * Calculates:
 * < W , div(f) >_boundary on the BOUNDARY. That is \f$ <W, \nabla \ dot f> \f$
 * Where:
 * W(x, y) is the shape function.
 * f(x, y) is the know function (RHS).
 * This is a *right* hand side term.
 * This is a *boundary* term.
 */
void u2d_ProblemCalculator::addW__divG_surf_right(u2d_BoundaryGaussPoints *bgp,
        u2d_NeumannCondition *cond, dense2D<double> field) {
    /* FIX ME, THE Iso Element should be deducted from BoundaryGaussPoints */
    /* Create an isoElement to recover weights from it */
    u2d_IsoElement1DP12Points isoElement;

    int m;
    for (int index = 0; index < cond->getNuemannElementNumber(); index++) {
        m = cond->neumann_elements(index) - 1; // Index of the element, because of the -1
        for (int gp = 0; gp < isoElement.gauss_points_n; gp++) {

            dense_vector<double> field_val(mesh->getSpatialDimension(), 0.0);
            for (int k = 0; k < mesh->getSpatialDimension(); k++) {
                for (int i = 0; i < BOUNDARY_ELEMENT_SIZE; i++) {
                    /* We use jjb_s here because the values in source have been generated in
                     * respect to the surface numeration of nodes. If source was numbered in
                     * respect to domain numeraton we would need to use jjb instead
                     */
                    field_val(k) += field(k, mesh->jjb_s(i, m) - 1) * isoElement.w(i, gp);
                }
            }
            /* Now field_val is ready and filled with values */

            double coef = 0;
            for (int i = 0; i < BOUNDARY_ELEMENT_SIZE; i++) {
                coef = isoElement.w(i, gp) * bgp->jacobian_weight(m);
                // WARNING: 2D dependent code
                double term = coef * (field_val[0] * bgp->norm_x[m] + field_val[1] * bgp->norm_y[m]);
                problem->addRHSTerm(mesh->jjb(i, m) - 1, term);
            }
        }
    }
}