/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on March 1, 2010, 10:20 AM
 */

#ifndef _U2D_SOLVER_H
#define	_U2D_SOLVER_H

#define MAX_ITER 500
#define MAX_ERR 1.e-6

#include "u2d_Problem.h"
#include "boost/numeric/mtl/mtl.hpp"
#include "boost/numeric/itl/itl.hpp"

using namespace mtl;
using namespace itl;

class u2d_Solver {
public:
    //u2d_Solver();
    //u2d_Solver(const u2d_Solver& orig);
    //virtual ~u2d_Solver();
    virtual dense_vector<double> getSolution() = 0;
    virtual size_t getSize() = 0;
    virtual dense_vector<double> solve() = 0;
private:

};

/**
 * This is a template for linear problems that should work with all the matrix
 * formats supported by MTL4. It is just a easy way of generating a subclass
 * of u2d_Solver with the right matrix type in it. The most basic type is
 * naturally one with a dense2D matrix type like:
 * u2d_LinearSolver<dense2D<double> >.
 */
template <typename Matrix>
class u2d_LinearSolver : public u2d_Solver {
public:
    u2d_LinearSolver(u2d_LinearProblem<Matrix> *_problem);
    u2d_LinearSolver(const u2d_LinearProblem<Matrix>& orig);
    virtual ~u2d_LinearSolver();
    /* Getters & Setters */
    dense_vector<double> getSolution();
    size_t getSize();
    void setMaxIter(int _maxiter);
    int getMaxIter();
    void setMaxErr(double _maxerr);
    double getMaxErr();
    /* Other methods */
    dense_vector<double> solve();

private:
    int maxiter;
    double maxerr;
    u2d_LinearProblem<Matrix> *problem;
    size_t size;
    dense_vector<double> solution;

};

template <typename Matrix>
u2d_LinearSolver<Matrix>::u2d_LinearSolver(u2d_LinearProblem<Matrix> *_problem) :
maxiter(MAX_ITER),
maxerr(MAX_ERR),
problem(_problem),
size(_problem->getSize()),
solution(_problem->getSize(), 0.0) {

}

template <typename Matrix>
u2d_LinearSolver<Matrix>::~u2d_LinearSolver() {
}

template <typename Matrix>
dense_vector<double> u2d_LinearSolver<Matrix>::solve() {
    itl::pc::ilu_0<Matrix> P(problem->A);
    itl::noisy_iteration<double> iter(problem->b, maxiter, maxerr);
    itl::bicgstab(problem->A, solution, problem->b, P, iter);
    return solution;
}

template <typename Matrix>
dense_vector<double> u2d_LinearSolver<Matrix>::getSolution() {
    return solution;
}

template <typename Matrix>
size_t u2d_LinearSolver<Matrix>::getSize() {
    return size;
}

/**
 * Set the maximum number of iterations. If a direct method i used this value
 * is simply ignored.
 * Default is MAX_ITER.
 */
template <typename Matrix>
void u2d_LinearSolver<Matrix>::setMaxIter(int _maxiter) {
    maxiter = _maxiter;
}

/**
 * Gets the current set number of iterations for iterative methods.
 * Default is MAX_ITER.
 */
template <typename Matrix>
int u2d_LinearSolver<Matrix>::getMaxIter() {
    return maxiter;
}

/**
 * Set the maximum error.
 * Default is MAX_ERR.
 */
template <typename Matrix>
void u2d_LinearSolver<Matrix>::setMaxErr(double _maxerr) {
    maxerr = _maxerr;
}

/**
 * Gets the current set maximum error.
 * Default is MAX_ERR.
 */
template <typename Matrix>
double u2d_LinearSolver<Matrix>::getMaxErr() {
    return maxerr;
}

#endif	/* _U2D_SOLVER_H */

