/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on March 8, 2010, 7:25 PM
 */


#include "u2d_Helmholtz.h"

/* For solver */
#include "u2d_mesh.h"
#include "u2d_load_umesh2d.h"
#include "u2d_IsoElement2DP13Points.h"
#include "u2d_IsoElement1DP12Points.h"
#include "u2d_DirichletCondition.h"
#include "u2d_NeumannCondition.h"
#include "u2d_GaussPoints.h"
#include "u2d_Problem.h"
#include "u2d_ProblemCalculator.h"
#include "u2d_Constants.h"
#include "u2d_Solver.h"
#include "solver/u2d_Constants.h"
#include "solver/u2d_Solver.h"
#include "file_output/u2d_PlotmtvOutput.h"
#include "solver/u2d_NeumannCondition.h"

/**
 * Solve the Helmholtz problem, that is:
 *
 *  [ - D.(k(r)D )  +  v(r).D  +  gamma ] u  =  source(r)
 *
 * The boundary conditions can be both Dirichlet and Robin, that is:
 *
 *    |
 *  u |    =  a(s)
 *    |S_D
 * 
 *     /                  du  \ |
 *    | alpha u  +  k(r)----   ||     =  b(s)
 *     \                  dn  / |S_R
 *
 * Alpha is called alpha, and determines if the boundary datum is of Neumann or
 * Robin type.
 * Gamma is called gamma.
 * K(r) is called thermal conductivity.
 * V(r) is called advection.
 * Source(r) is called source.
 * @remark This constructor expects that the mesh is already stored in the
 * private variable, so this should not be called from outside. It is a private
 * constructor.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
u2d_Helmholtz::u2d_Helmholtz() :
alpha(0),
gamma(0),
//source(mesh->getDomainNodeNumber(), 0.0),
source(0, 0.0),
avector(0, 0.0),
bvector(0, 0.0),
//kvector(mesh->getDomainNodeNumber(), 0.0),
kvector(0, 0.0),
bfield(0, 0),
//vfield(mesh->getSpatialDimension(), mesh->getDomainNodeNumber())
vfield(0, 0) {

}

/**
 * Creates an Helmholtz problem give path and the project name where
 * the mesh files are located. The files must be in umesh2d format named
 * grid.problemname anf nodes.problemname that is the standard nameing for
 * files created with umesh2d.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
u2d_Helmholtz::u2d_Helmholtz(string meshdir, string problemname) {
   u2d_Load_Umesh2d mesh_loader(meshdir, problemname, true);
   this->mesh = mesh_loader.getMesh();
   u2d_Helmholtz();
}

/**
 * Creates an Helmholtz problem given a pointer to an already loaded mesh.
 * The mesh is copied internaly and so the passed pointer can be deallocated.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
u2d_Helmholtz::u2d_Helmholtz(u2d_Mesh *_mesh) {
//    mesh = new u2d_Mesh(*_mesh);
    u2d_Helmholtz();
}

u2d_Helmholtz::u2d_Helmholtz(const u2d_Helmholtz& orig) {
}

u2d_Helmholtz::~u2d_Helmholtz() {

}

void u2d_Helmholtz::run() {
    dgp = ie2d.applyTo(mesh);
    bgp = ie1d.applyTo(mesh);
    dc.generateNodes(mesh);
    nc.calculateElements(mesh);
    problem = new u2d_LinearProblem<dense2D<double> >(mesh->getDomainNodeNumber());
    calculator = new u2d_ProblemCalculator(problem, mesh);
    /* LHS (Matrix) terms */
    calculator->addDW__AlphaDG_left(dgp, kvector);
    calculator->addW__AlphaDG_left(dgp, vfield);
    calculator->addAlphaW__G_left(dgp, gamma);
    calculator->addAlphaW__G_surf_left(bgp, &nc, alpha);
    calculator->forceDirichlet_left(&dc);
    /* RHS terms */
    calculator->addW__G_right(dgp, source);
    calculator->addW__G_surf_right(bgp, &nc, bvector);
    calculator->addW__divG_surf_right(bgp, &nc, bfield);
    calculator->forceDirichlet_right(&dc, avector);
    /* A solutor.. and solve :) */
    u2d_LinearSolver<dense2D<double> > solver(problem);
    solver.solve();
    /* Output to screen */
    cout << solver.getSolution() << endl;
    /* Output to plotmtv format TODO, REMOVE ME ASAP WE CAN DISPLAY SOLUTION INTERNALLY */
    u2d_PlotmtvOutput plotmtvout("solution", mesh->rr, mesh->jjd, solver.getSolution());
    plotmtvout.output();
}

void u2d_Helmholtz::setSideNeumann(int side_id) {
    nc.setSide(side_id);
}

void u2d_Helmholtz::setSideDirichlet(int side_id) {
    dc.setSide(side_id);
}

void u2d_Helmholtz::setGamma(double _gamma) {
    gamma = _gamma;
}

void u2d_Helmholtz::setAlpha(double _alpha) {
    alpha = _alpha;
}

/**
 * Set the vector of Dirichlet codition.
 * Set the vector a(r) that is the Dirichlet condition on the Dirichlet boundary,
 * consequentely this must have lenght equal to the number of dirichlet nodes of
 * the boundary. The nodes are numbered by boundary numeration and ordered as
 * the unique node set that is generated by the u2d_DirichletCondition class.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setAVector(dense_vector<double> _avector) {
    /* check that _avector is the same size as dirichlet codition sides */
    /* eventually resize avector to the appropriate size */
    avector = _avector;
}

/**
 * Set the vector of Robin codition.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setBVector(dense_vector<double> _bvector) {
    /* add excpetion if wrong size size of boundary nodes */
    bvector = _bvector;
}

/**
 * Set the Thermal (K) Conductivity scalar field.
 * The field is represented as a vector where each element is the value of the
 * thermal conductivity in the corresponding mesh node. Consequently this vector
 * has lenght equal to the number of domain nodes.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setKVector(dense_vector<double> _kvector) {

    kvector = _kvector;
}

/**
 * Set the vector field of Robin codition.
 * The filed is represented as N row vectors one above the other. That is:
 * It is a NxM matrix where N is the number of spatial dimentions the problem
 * has and M is the number of domain nodes.
 * I.e. in 2D:
 *
 *   V = | 0.345   0.762   ...   0.997 |   value of X
 *       | 0.234   0.748   ...   0.134 |   value of Y
 *         ^^^^^   ^^^^^         ^^^^^
 * Domain Node 1   D.Node 2      D.Node M
 * 
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setBField(dense2D<double> _bfield) {
    /* add excpetion if wrong size ????size of boundary nodes */
    bfield = _bfield;
}

/**
 * Set the Advection velocity vector field.
 * The filed is represented as N row vectors one above the other. That is:
 * It is a NxM matrix where N is the number of spatial dimentions the problem
 * has and M is the number of domain nodes.
 * I.e. in 2D:
 *
 *   V = | 0.345   0.762   ...   0.997 |   value of X
 *       | 0.234   0.748   ...   0.134 |   value of Y
 *         ^^^^^   ^^^^^         ^^^^^
 * Domain Node 1   D.Node 2      D.Node M
 *
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setVField(dense2D<double> _vfield) {
    /* add excpetion if wrong size */
    vfield = _vfield;
}

/**
 * Sets the source scalar field
 * The field is represented as a vector where each element is the intensity of
 * the source in the corresponding mesh node. Consequently this vector
 * has lenght equal to the number of domain nodes.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setSource(dense_vector<double> _source) {
    /* add excpetion if wrong size */
    source = _source;
}

/**
 * Run a self contained test for a Helmholtz problem on a square with known
 * analytical solution. The Difference from exact solution is printed on screen.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::SquareTest() {
    /* Load the mesh */
    u2d_Load_Umesh2d mesh_loader("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square", "square_c", true);
    u2d_Mesh *mesh = mesh_loader.getMesh();
    /* Calculate the iso elements and the domain and boundary gauss points */
    u2d_IsoElement2DP13Points ie2d;
    u2d_IsoElement1DP12Points ie1d;
    u2d_DomainGaussPoints *dgp = ie2d.applyTo(mesh);
    u2d_BoundaryGaussPoints *bgp = ie1d.applyTo(mesh);
    /* Calculate Dirichlet Boundary Condition */
    u2d_DirichletCondition dc;
    dc.setSide(1);
    dc.setSide(2);
    dc.setSide(3);
    dc.generateNodes(mesh);
    /* Calculate Von Neumann Condition */
    u2d_NeumannCondition nc;
    nc.setSide(4);
    nc.calculateElements(mesh);


    /* Create a problem */
    u2d_LinearProblem<dense2D<double> > problem(mesh->getDomainNodeNumber());
    /* Calculate various terms of the problem */
    u2d_ProblemCalculator calculator(&problem, mesh);
    /* LHS (Matrix) terms */
    calculator.addDW__AlphaDG_left(dgp, u2d_Constants::thermalConductivity(mesh->rr));
    calculator.addW__AlphaDG_left(dgp, u2d_Constants::advection(mesh->rr));
    double gamma = 10.0;
    calculator.addAlphaW__G_left(dgp, gamma);
    double alpha = 2.0;
    calculator.addAlphaW__G_surf_left(bgp, &nc, alpha);
    calculator.forceDirichlet_left(&dc);
    /* RHS terms */
    calculator.addW__G_right(dgp, u2d_Constants::source(mesh->rr, gamma));
    /* Build a matrix with coordinates of boundary nodes */
    dense2D<double> rrb(mesh->getSpatialDimension(), mesh->getBoundaryNodeNumber());
    for (int i = 0; i < mesh->getBoundaryNodeNumber(); i++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rrb(j, i) = mesh->rr(j, mesh->jjb_sd(i) - 1);
        }
    }
    /* Prepare the solution for addW__G_right on boundary */
    dense_vector<double> pyt_alpha(mesh->getBoundaryNodeNumber());
    pyt_alpha = alpha * u2d_Constants::pythagoras(rrb);
    calculator.addW__G_surf_right(bgp, &nc, pyt_alpha);
    /* Prepare the field for addW__divG_right on the boundary */
    dense_vector<double> thermal_k(num_cols(rrb));
    thermal_k = u2d_Constants::thermalConductivity(rrb);
    dense2D<double> spread_thermal_k(2, num_rows(thermal_k));
    spread_thermal_k[0][mtl::irange()] = trans(thermal_k);
    spread_thermal_k[1][mtl::irange()] = trans(thermal_k);
    dense2D<double> grad_pyt(mesh->getSpatialDimension(), num_cols(rrb));
    grad_pyt = u2d_Constants::gradPythagoras(rrb);
    dense2D<double> field(mesh->getSpatialDimension(), num_cols(rrb));
    field = ele_prod(spread_thermal_k, grad_pyt);
    calculator.addW__divG_surf_right(bgp, &nc, field);
    /* Force the boundary condition on the RHS terms */
    /* Build a matrix with coordinates of boundary dirichlet nodes */
    dense2D<double> rr_dirichlet(mesh->getSpatialDimension(), dc.dirichlet_unique_nodes.size());
    std::set<int>::iterator it;
    int rr_d_index = 0;
    for (it = dc.dirichlet_unique_nodes.begin();
            it != dc.dirichlet_unique_nodes.end();
            it++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rr_dirichlet(j, rr_d_index) = mesh->rr(j, *it - 1);
        }
        ++rr_d_index;
    }
    calculator.forceDirichlet_right(&dc, u2d_Constants::pythagoras(rr_dirichlet));
    /* Let's try a solutor */
    u2d_LinearSolver<dense2D<double> > solver(&problem);
    solver.solve();


    /* Debug output */
    // cout << num_rows(rrb) << ", " << num_cols(rrb) << endl;
    // cout << problem.A << endl;
    // cout << problem.b << endl;
    cout << solver.getSolution() << endl;
    cout << u2d_Constants::pythagoras(mesh->rr) << endl; // solution in a square
    dense_vector<double> sol_error(num_rows(solver.getSolution()), 0.0);
    sol_error = solver.getSolution() - u2d_Constants::pythagoras(mesh->rr);
    cout << sol_error << endl;

    /* Output a plot file */
    u2d_PlotmtvOutput plotmtvout("solution", mesh->rr, mesh->jjd, solver.getSolution());
    u2d_PlotmtvOutput plotmtvout2("solution_exact", mesh->rr, mesh->jjd, u2d_Constants::pythagoras(mesh->rr));
    plotmtvout.output();
    plotmtvout2.output();
}


/**
 * Run a self contained test for a Helmholtz problem on a square with known
 * analytical solution. The Difference from exact solution is printed on screen.
 * This is the same as SquareTest just tailored with more debug output to
 * examinate each term indipentently.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::SquareTest_debug() {
    /* Load the mesh */
    u2d_Load_Umesh2d::u2d_Load_Umesh2d mesh_loader("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square", "square_c", true);
    u2d_Mesh *mesh = mesh_loader.getMesh();
    /* Calculate the iso elements and the domain and boundary gauss points */
    u2d_IsoElement2DP13Points::u2d_IsoElement2DP13Points ie2d;
    u2d_IsoElement1DP12Points::u2d_IsoElement1DP12Points ie1d;
    u2d_DomainGaussPoints *dgp = ie2d.applyTo(mesh);
    u2d_BoundaryGaussPoints *bgp = ie1d.applyTo(mesh);
    /* Calculate Dirichlet Boundary Condition */
    u2d_DirichletCondition::u2d_DirichletCondition dc;
    dc.setSide(1);
    dc.setSide(2);
    dc.setSide(3);
    dc.generateNodes(mesh);
    /* Calculate Dirichlet Von Neumann Condition */
    u2d_NeumannCondition::u2d_NeumannCondition nc;
    nc.setSide(4);
    nc.calculateElements(mesh);


    //u2d_Constants::thermalConductivity(mesh->rr);  OK!!!
    //u2d_Constants::advection(mesh->rr); OK!!!
    //u2d_Constants::pythagoras(mesh->rr); OK!!!
    //u2d_Constants::gradThermalConductivity(mesh->rr); OK!!!
    //u2d_Constants::gradPythagoras(mesh->rr); OK!!!
    //u2d_Constants::source(mesh->rr, gamma); OK!!!
    //u2d_Constants::laplacianSolution(mesh->rr); OK!!!
    // field is OK!!!


    /* To check dirichlet */
    //    dense2D<double> rr_dirichlet2(mesh->getSpatialDimension(), dc.dirichlet_unique_nodes.size());
    //    std::set<int>::iterator it2;
    //    int rr_d_index2 = 0;
    //    for (it2 = dc.dirichlet_unique_nodes.begin();
    //            it2 != dc.dirichlet_unique_nodes.end();
    //            it2++) {
    //        printf("%d\n", *it2);
    //        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
    //            rr_dirichlet2(j, rr_d_index2) = mesh->rr(j, *it2 - 1);
    //        }
    //        ++rr_d_index2;
    //    }
    /* To check field */
    //    dense2D<double> rrb2(mesh->getSpatialDimension(), mesh->getBoundaryNodeNumber());
    //    for (int i = 0; i < mesh->getBoundaryNodeNumber(); i++) {
    //        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
    //            rrb2(j, i) = mesh->rr(j, mesh->jjb_sd(i) - 1);
    //        }
    //    }
    //    dense_vector<double> thermal_k2(num_cols(rrb2));
    //    thermal_k2 = u2d_Constants::thermalConductivity(rrb2);
    //    dense2D<double> spread_thermal_k2(2, num_rows(thermal_k2));
    //    spread_thermal_k2[0][mtl::iall] = trans(thermal_k2);
    //    spread_thermal_k2[1][mtl::iall] = trans(thermal_k2);
    //    dense2D<double> grad_pyt2(mesh->getSpatialDimension(), num_cols(rrb2));
    //    grad_pyt2 = u2d_Constants::gradPythagoras(rrb2);
    //    dense2D<double> field2(mesh->getSpatialDimension(), num_cols(rrb2));
    //    field2 = ele_prod(spread_thermal_k2, grad_pyt2);


    //    dense_vector<double> vect = u2d_Constants::laplacianSolution(mesh->rr);
    //    for (int i = 0; i < num_rows(vect); i++) {
    //        printf("%f\n", vect(i));
    //    }

    //    dense2D<double> mat = field2;
    //    for (int i = 0; i < num_rows(mat); i++) {
    //        for (int j = 0; j < num_cols(mat); j++) {
    //            printf("%f \t %f\n", mat(0, j), mat(1, j));
    //        }
    //    }

    /* To check RHS and LHS terms */
    u2d_LinearProblem<dense2D<double> > problem(mesh->getDomainNodeNumber());
    u2d_ProblemCalculator calculator(&problem, mesh);

    /* To check addDW__AlphaDG_left */
    calculator.addDW__AlphaDG_left(dgp, u2d_Constants::thermalConductivity(mesh->rr));

    /* To check addW__AlphaDG_left */
    calculator.addW__AlphaDG_left(dgp, u2d_Constants::advection(mesh->rr));

    /* To check addAlphaW__G_left */
    double gamma = 10.0;
    calculator.addAlphaW__G_left(dgp, gamma);

    /* To check addAlphaW__G_surf_left */
    double alpha = 2.0;
    calculator.addAlphaW__G_surf_left(bgp, &nc, alpha);

    /* To check forceDirichlet_left */
    calculator.forceDirichlet_left(&dc);

    /* To check addW__G_right */
    //double gamma = 10.0;
    calculator.addW__G_right(dgp, u2d_Constants::source(mesh->rr, gamma));

    /* To check addW__G_surf_right */
    //double alpha = 2.0;
    dense2D<double> rrb(mesh->getSpatialDimension(), mesh->getBoundaryNodeNumber());
    for (int i = 0; i < mesh->getBoundaryNodeNumber(); i++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rrb(j, i) = mesh->rr(j, mesh->jjb_sd(i) - 1);
        }
    }
    ///* Prepare the solution for addW__G_right on boundary */
    dense_vector<double> pyt_alpha(mesh->getBoundaryNodeNumber());
    pyt_alpha = alpha * u2d_Constants::pythagoras(rrb);
    calculator.addW__G_surf_right(bgp, &nc, pyt_alpha);

    /* To check addW__divG_surf_right */
    /* Prepare the field for addW__divG_right on the boundary */
    //dense2D<double> rrb(mesh->getSpatialDimension(), mesh->getBoundaryNodeNumber());
    //for (int i = 0; i < mesh->getBoundaryNodeNumber(); i++) {
    //    for (int j = 0; j < mesh->getSpatialDimension(); j++) {
    //        rrb(j, i) = mesh->rr(j, mesh->jjb_sd(i) - 1);
    //    }
    //}
    dense_vector<double> thermal_k(num_cols(rrb));
    thermal_k = u2d_Constants::thermalConductivity(rrb);
    dense2D<double> spread_thermal_k(2, num_rows(thermal_k));
    spread_thermal_k[0][mtl::irange()] = trans(thermal_k);
    spread_thermal_k[1][mtl::irange()] = trans(thermal_k);
    dense2D<double> grad_pyt(mesh->getSpatialDimension(), num_cols(rrb));
    grad_pyt = u2d_Constants::gradPythagoras(rrb);
    dense2D<double> field(mesh->getSpatialDimension(), num_cols(rrb));
    field = ele_prod(spread_thermal_k, grad_pyt);
    calculator.addW__divG_surf_right(bgp, &nc, field);

    /* To check forceDirichlet_right */
    /* Build a matrix with coordinates of boundary dirichlet nodes */
    dense2D<double> rr_dirichlet(mesh->getSpatialDimension(), dc.dirichlet_unique_nodes.size());
    std::set<int>::iterator it;
    int rr_d_index = 0;
    for (it = dc.dirichlet_unique_nodes.begin();
            it != dc.dirichlet_unique_nodes.end();
            it++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rr_dirichlet(j, rr_d_index) = mesh->rr(j, *it - 1);
        }
        ++rr_d_index;
    }
    calculator.forceDirichlet_right(&dc, u2d_Constants::pythagoras(rr_dirichlet));


#define vectorout 0
#define matrixout 1


#ifdef vectorout
    dense_vector<double> vect = problem.b;
    double total = 0.0;
    for (int i = 0; i < num_rows(vect); i++) {
        printf("%i \t %e\n", i, vect(i));
        total += vect(i);
    }
    printf("%e\n", total);
#endif

#ifdef matrixout
    dense2D<double> mat = problem.A;
    double total_mat = 0.0;
    for (int i = 0; i < num_rows(mat); i++) {
        for (int j = 0; j < num_cols(mat); j++) {
            if (mat(i, j) != 0) {
                printf("%d \t %d \t %e\n", i, j, mat(i, j));
                total_mat += mat(i,j);
            }
        }
    }
    printf("%e\n", total_mat);
#endif
    //cout << problem.A << endl;

    return;

}